/* Copyright 1994 GROUPE BULL -- See license conditions in file COPYRIGHT */
/* $Id: KnvasP.h,v 1.5.1.2 95/05/23 11:54:41 leon Exp $ */
#ifndef __KNVASP_H
#define __KNVASP_H

/* set default to 3D LOOK */
#ifndef NO_RELIEF
#define LOOK3D
#endif

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <Kn/Knvas.h>
#include <Xm/Xm.h>
#include <Xm/ScrolledW.h>
#include <Xm/ScrollBar.h>
#if (XmVersion >= 1002)
#include <Xm/ManagerP.h>
#else
#include <Xm/XmP.h>
#endif /* (XmVersion >= 1002) */



#ifdef __cplusplus
extern "C" {
#endif


typedef struct {
    int empty;
} KnvasClassPart;

typedef struct _KnvasClassRec {
    CoreClassPart	core_class;
    CompositeClassPart	composite_class;
    ConstraintClassPart		constraint_class;
    XmManagerClassPart		manager_class;
    KnvasClassPart	knvas_class;
} KnvasClassRec;

enum InteractionCursor {
    KnCENTER_CURSOR = 0,
    KnTOP_LEFT_CURSOR = 1,
    KnTOP_CURSOR = 2,
    KnTOP_RIGHT_CURSOR = 3,
    KnBOTTOM_LEFT_CURSOR = 4,
    KnBOTTOM_CURSOR = 5,
    KnBOTTOM_RIGHT_CURSOR = 6,
    KnLEFT_CURSOR = 7,
    KnRIGHT_CURSOR = 8,
    KnERROR_CURSOR = 9
};

typedef struct _KnvasPart{
    /* resources (public fields) */
    XtCallbackList interactorCallback; /* called by any interactor  function */
    XtCallbackList knvasCallback; /* called by globals interactors */
    Boolean pixmapBuffering;
    Boolean autoUpdate;

    /* private fields */
    Transformer t;		/* view transformer */
    XtValueMask smask;		/* OR of all the kno masks */
    KnTag defaultTag;		/* This tag will be set on every newly created
				   object */		       
    KlList layers;		/* a view consists in a list of layers */
    KlList tags;		/* each view must have a palette of tags */
    KlList interactors;		/* each view must have a palette of tags */
    KlList sobjs;		/* selected objects */
    KnRegion clear;		/* build the region to erase */
    KlList delete;		/* obj to be deleted */
    KlList clear_tags;
    Pixmap pixmap;		/* pixmap Id when pixmapBuffering is on */
    XRectangle psize;		/* pixmap size when pixmap buffering is on */
    GC pixgc;			/* gc used to copy from pixmap to window */
    GC ghostgc;
    GC selectgc;
    GC maskgc1;			/* used to draw clipped - depth 1 */
    GC maskgc;			/* used to draw clipped */
    int kindOfMove;
    Boolean dispatched;		/* XEvent dispatched by an interactor ? */
    Boolean in_dispatch;	/* are we inside a dispatch proc ? */
    Cursor cursors[10];		/* cursors used by interactors */
    KnO grab;			/* object grabbing events */
    KlList gobjs;		/* ghost list (objects that are under
				   interaction) */
    Transformer *gt;		/* each gobj has a global transformer */
    KnO target;			/* Image that is currently considered as this
				   widget */
    Position lastx, lasty;	/* position of last move */
    Position firstx, firsty;
#ifdef LOOK3D
    Pixmap stipple;		/* we need to build a stipple / display */
#endif
    int xmin, xmax;		/* horizontal min/max values of scrolling */
    int ymin, ymax;		/* vertical min/max values of scrolling */
    int hval, vval;
    float zoomx, zoomy;		/* keep trace of current zoom factor */
    XmScrollBarWidget hsb;	/* the horizontal scrollbar */
    XmScrollBarWidget vsb;	/* the vertical scrollbar */
    XmScrolledWindowWidget sw;	/* the parent scrolled window */
} KnvasPart;


typedef struct _KnvasRec {
    CorePart		core;
    CompositePart	composite;
    ConstraintPart      constraint;
    XmManagerPart       manager;
    KnvasPart	        knvas;
} KnvasRec;

extern KnvasClassRec knvasClassRec;


#ifdef __cplusplus
} /* Close scope of 'extern "C"' */
#endif


#endif /* __KNVASP_H */
