/* Copyright 1994 GROUPE BULL -- See license conditions in file COPYRIGHT */
/* $Id: group.c,v 1.8.1.3 95/05/23 11:54:08 leon Exp $ */


#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include "misc.h"
#include "group.h"


/* type desc */
KlType KnGroupClass;



void
KnGroupInit(KnGroup group, Widget knvas, KnO parent, Position x, Position y)
{
    KnOInit((KnO)group, knvas, parent, x, y);
    group->children = KlListNMake(0);
    KlIncRef(group->children);
    group->t = TransformerMake();
    TransformerTranslate(group->t, x, y);
}


void
KnGroupFree(KnGroup group)
{
    KlDecRef(group->children);
    KnOFree((KnO)group);
}


/* allocates storage space for a new KnGroup object */
KnO
KnCreateGroup(Widget knvas, KnO parent, Position x, Position y)
{  
    KnGroup group;
    group =  (KnGroup)KlOMake(KnGroupClass);
    KnGroupInit(group, knvas, parent, x, y);
    return (KnO)group;
}





KnO
KnGroupAdd(KnGroup g, KnO kno)
{
    KlListAppend(g->children, (KlO)kno);

    return kno;
}



void
KnGroupRemove(KnGroup g, KnO kno)
{
    Cardinal n;

    for(n = 0; n < g->children->size; n++) {
	if((KnO)g->children->list[n] == kno) {
	    KlListDeletePos(g->children, n);
	}
    }
}




void
KnGroupDraw(KnGroup me, DrawContext ct, Transformer tf) 
{
    Transformer t;   
    Cardinal i;
    KnO child;
    Boolean need_free;
    
    for(i = 0; i < me->children->size; i++) {
	child = (KnO)me->children->list[i];
	if(child && KnIsMapped(child)) {
	    /* draw only mapped objects */
	    t = tf;
	    need_free = False;
	    if(t) {
		if(child->t) {
		    t = TransformerMakeCopy(tf);
		    TransformerPremultiply(t, child->t);
		    if(TransformerIdentity(t)) {
			free(t);
			t = NULL;
		    }
		    else
			need_free = True;
		}
	    }
	    else {
		t = child->t;
	    }
	    KlSend_draw(child, ct, t);
	    if(need_free) {
		free(t);
		need_free = False;
	    }
	}
    }
}

void
KnGroupClear(KnGroup me, Widget knvas)
{
    KnO child;
    int i;


    for(i = 0; i < me->children->size; i++) {
	child = (KnO)me->children->list[i];
	if(child && KnIsMapped(child)) {
	    KlSend_clear(child, knvas);
	}
    }
}




KnO
KnGroupContains(KnGroup group, Widget knvas, Int x, Int y)
{
    KnO res, child;
    Position rx = x, ry = y;
    int i;

    if(group->t) 
	TransformerInvTransform(group->t, x, y, &rx, &ry);
    else {
	rx -= group->x;
	ry -= group->y;
    }
    for(i = group->children->size-1; i >= 0; i--) {
	child = (KnO)group->children->list[i];
	if(child && KnIsMapped(child)) {
	    res = (KnO)KlSend_contains(child, knvas,
				       (Int)rx, (Int)ry);
	    if(res)
		return res;
	}
    }
    return NULL;
}





void
KnGroupBoundingBox(KnO kno, Widget view, XRectangle *box,
	       XPoint *points, Cardinal n)
{
    if(kno->t) {
	TransformerTransformPointList(kno->t, points, n, points);
    }
    KlSend_boundingbox(kno->parent, view, box, points, n);
}



/* XtDispatch this will always return True, because the event filter, is ok.
   should install the translation's event filter in the widget */
Int
KnGroupDispatch(KnGroup group, Widget w, XEvent *event, Int x, Int y)
{
    Position rx = x, ry = y;
    int i;
    KnO child;
    XtTMRec tm;
    KnvasWidget cw = (KnvasWidget) w;

    if( ! KnIsSensitive(group)) return (Int)False;
    
    if(group->interactor) {
	if(KlSend_contains(group, w, (Int)x, (Int)y)) {
	    cw->knvas.target = (KnO)group;

	    tm = w->core.tm;
	    w->core.tm = group->interactor->tm;
	    
	    _XtTranslateEvent(w, event);
	    
	    group->interactor->tm = w->core.tm;
	    w->core.tm = tm;
	    
	    return (Int)(cw->knvas.dispatched);
	}
    }
    if(group->t) 
	TransformerInvTransform(group->t, x, y, &rx, &ry);
    for(i = group->children->size - 1; i >=0; i--) {
	child = (KnO)group->children->list[i];
	if(KnIsMapped(child) && KnIsSensitive(child)) {
	    if(KlSend_dispatch(child, w, event, (Int)rx, (Int)ry))
		return (Int)(cw->knvas.dispatched);
	}
    }
    return (Int)False;
}



/* declare new   selectors
 */
void
KnGroupImport()
{
	
}



/* class initializations for KnGroup objects
 * 
 * returns: 
 */
void
KnGroupClassInitialize()
{
     /* declare  type */
    KlDeclareSubType(&KnGroupClass, "KnGroup", KnOClass, sizeof(struct KnGroupStruct));

    /* declare methods */
    KlDeclareMethod(KnGroupClass, KlSelMake, KnCreateGroup);
    KlDeclareMethod(KnGroupClass, KlSelFree, KnGroupFree);
    KlDeclareMethod(KnGroupClass, KnOSelDraw, KnGroupDraw);
    KlDeclareMethod(KnGroupClass, KnOSelClear, KnGroupClear);
    KlDeclareMethod(KnGroupClass, KnOSelContains, KnGroupContains);
    KlDeclareMethod(KnGroupClass, KnOSelDispatch, KnGroupDispatch);
    KlDeclareMethod(KnGroupClass, KnOSelBoundingBox, KnGroupBoundingBox);


}
