static char rcsid[]="$Id: hash.c,v 1.2.1.1 95/04/20 18:10:02 leon Exp $";
/* Copyright 1989-93 GROUPE BULL -- See license conditions in file COPYRIGHT */

/***************************************************************************
 * khash.c
 *
 * This is a simple generic package for hash tables. According to the
 * mode of the hash table, it will acccept (or not) items with the same
 * key.
 *
 * static:
 *
 * externs:
 * HashtableMake  : make a new hashtable
 * HashtableAdd   : add the given entry in hashtable
 * HashtableFind
 * HashtableFindNext 
 * HashtableMap:
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Intrinsic.h>	/* for XtNewString */

#include "hash.h"


/*
 * A few std cmp/hash function sets
 */

unsigned long OsHashCString (char *cstr) {
    unsigned long ret = 0;
    unsigned long ctr = 0;

    while (*(char*)cstr) {
        ret ^= *(char*)cstr++ << ctr;
        ctr = (ctr + 1) % sizeof (void *);
    }
    return (ret) ;
}

int OsCmpCString (char *s1, char *s2) {
    return (strcmp (s1, s2)) ;
}


/***************************************************************************
 * HashtableMake: make a new hashtable.
 *
 * RETURN: Hashtable, exit on error
 ***************************************************************************/

Hashtable HashtableMake (size)
    unsigned int  size ;
{
    Hashtable hashtable ;

    if ( ( ! ( hashtable = (Hashtable) malloc (sizeof (HashtableRec))))
	|| ( ! (hashtable->slots = (Hashslot *) malloc (sizeof (Hashslot *) * size))))
	
    {
	perror ("HashtableMake: malloc ") ;
	exit (1) ;
    }

    memset ((char *) hashtable->slots, 0, sizeof (Hashslot *) * size) ;
    hashtable->size         = size ;
    hashtable->item_count   = 0 ;
    hashtable->hash_fct     = OsHashCString ;
    hashtable->cmp_fct      = OsCmpCString ;
    hashtable->mode         = 0 ;
    hashtable->current_slot = (Hashslot) 0 ;

    return (hashtable) ;
}

/***************************************************************************
 * HashtableAdd: add the given item to hashtable
 * HashtableRemove: remove given item from hashtable.
 *
 * Slightly depends on the hashtable mode for Add. (Overide)
 * Removes compare item's adresses, not items key.
 *
 * RETURN: a pointer to the item (might != arg, because already in)
 ***************************************************************************/

void *HashtableAdd (hashtable, item, value)
    register Hashtable hashtable ;
    char *item;
    void               *value ;

{
    register Hashslot slot ;
    Hashslot          *position ;

/*
 * Get the correct slot, check for present :
 */
    position = &(hashtable->slots[(hashtable->hash_fct)(item) % hashtable->size]) ;
    slot     = *position ;
    if ( hashtable->mode & HASH_OVERRIDE )
    {
	while ( slot )
	{
	    if ( ! (hashtable->cmp_fct)(slot->item, item) )
		return (slot->item) ;
	    slot = slot->next ;
	}
    }
/*
 * We must add it :
 */
    if ( ! (slot = (Hashslot) malloc (sizeof (HashslotRec))) )
    {
	perror ("HashtableAdd, malloc ") ;
	exit (1) ;
    }

    (hashtable->item_count)++ ;
    slot->item = XtNewString(item) ;
    slot->value = value;
    slot->next = *position ;
    *position  = slot ;
    return (slot) ;
}

void *HashtableRemove (hashtable, item, value)
    Hashtable hashtable ;
    char *item;
    void      *value ;

{
    register Hashslot slot, *pslot ;
    unsigned long     hashval ;

    hashval = (hashtable->hash_fct) (item) % hashtable->size ;
    slot  = hashtable->slots[hashval] ;
    pslot = &(hashtable->slots[hashval]) ;
    while ( slot )
    {
	if ( ! strcmp(slot->item, item) )
	{
	    (hashtable->item_count)-- ;
	    *pslot = slot->next ;
	    free(slot->item);
	    free (slot) ;
	    return (item) ;
	}
	pslot = &(slot->next) ;
	slot  = slot->next ;
    }

    return ((void *) 0) ;
}

/***************************************************************************
 * HashtableFind: retrieve the given entry
 *
 * If the hashtable is not in override mode, set its internal lookup
 * pointer to the found item, so that FindNext can proceed.
 *
 * RETURN: the entry in hashtable
 ***************************************************************************/

void *HashtableFind (hashtable, item)
    register Hashtable hashtable ;
    char               *item ;

{
    register Hashslot slot ;

    slot = hashtable->slots[(hashtable->hash_fct)(item) % hashtable->size] ;
    while ( slot )
    {
	if ( ! (hashtable->cmp_fct)(slot->item, item) )
	{
	    if ( ! (hashtable->mode & HASH_OVERRIDE) )
		hashtable->current_slot = slot ;
	    return (slot->value) ;
	}
	slot = slot->next ;
    }
    
    return ((void *) 0) ;
}
