/* $Id: hash.h,v 1.2 95/03/24 15:35:56 leon Exp $ */
/* Copyright 1989-93 GROUPE BULL -- See license conditions in file COPYRIGHT */

/***************************************************************************
 * khash.h
 *
 * Simple hash table management functions.
 ***************************************************************************/



#ifndef _HASHTABLE_H_
#define _HASHTABLE_H_


#include <Kn/compat.h>

typedef struct _Hashslot {
    struct _Hashslot 	*next ;		/* collision handling */
    char *item;
    void		*value ;		/* item at this entry */
} HashslotRec, *Hashslot ;

typedef struct _Hashtable {
    unsigned long	(*hash_fct)() ;	/* hash function on those items */
    int			(*cmp_fct)() ;	/* compare function on those items */
    unsigned int	size ;		/* # of entries in table */
    unsigned int	item_count ;	/* # of items actually stored */
    unsigned int	mode ;		/* mode of the hashtable */
    Hashslot		*slots ;	/* table of slots */
    Hashslot		current_slot ;	/* current lookup value (FindNext) */
} HashtableRec, *Hashtable ;

/*
 * Various defines for the hashtable 'mode' field :
 */

#define HASH_OVERRIDE		1

/*
 * Macros to access slots :
 */

#define HashthableSize(h)	 ((h)->size)
#define HashtableCount(h)	 ((h)->item_count)

#define HashtableOverrideMode(h) ((h)->mode |= HASH_OVERRIDE)

/*
 * Prototypes :
 */

Hashtable HashtableMake ( unsigned int size );


void *HashtableAdd  (Hashtable hashtable, char *item, void *value) ;
void *HashtableRemove (Hashtable hashtable, char *item, void *value) ;
void *HashtableFind (Hashtable hashtable, char *item);



#endif /* _HASHTABLE_H_ */

