/* Copyright 1994 GROUPE BULL -- See license conditions in file COPYRIGHT */
/* $Id: interactor.c,v 1.6.1.2 95/04/28 16:20:14 leon Exp $ */
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include "interactor.h"
#include "select-move.h"
#include "resize.h"
#include "hash.h"
#include "cvt.h"


/* base file for creating a new klone type */

/* type desc */
KlType KnInteractorClass;




static XtActionsRec
actions[] = {
    {"Select", SelectProc},

    {"SelectDrag", SelectDragProc},
    {"SelectDrop", SelectDropProc},
    
    {"MoveDrag", MoveDragProc},
    {"MoveDrop", MoveDropProc},

    {"ResizeDrag", ResizeDragProc},
    {"ResizeDrop", ResizeDropProc},
};



static XtResource 
interactorResources [] = {
#define offset(field) XtOffset(KnInteractor, field)
  {
      XtNvisibleSelection, XtCVisibleSelection,
      XtRBoolean, sizeof(Boolean), offset(visibleSelection),
      XtRImmediate, (caddr_t)False
  },
  {
      XtNbutton, XtCButton,
      XtRInt, sizeof(int), offset(button),
      XtRImmediate, (caddr_t)KnLEFT
  }, 
  {
      XtNconstraints, XtCConstraints,
      XtRString, sizeof(String), offset(constraints),
      XtRImmediate, (caddr_t)0
  },
  {
      XtNvaryCursor, XtCVaryCursor,
      XtRBoolean, sizeof(Boolean), offset(varyCursor),
      XtRImmediate, (caddr_t)True
  },
  {
      XtNcallbackPolicy, XtCCallbackPolicy,
      XtRCallbackPolicy, sizeof(KnCallbackPolicy), offset(callbackPolicy),
      XtRImmediate, (caddr_t)KnCALLBACK_ALWAYS
  }, 
  
#undef offset
};


static Boolean
CvtStringToCallbackPolicy(Display *dpy, XrmValue *args, Cardinal *num_args, 
			  XrmValue *from, XrmValue *to,
			  XtPointer *converter_data)
{
    int i;
    char s[32];
    static struct _names {
	char *name;
	int style;
    } names[] = {
	{"callback_never", KnCALLBACK_NEVER},
	{"callback_on_pick", KnCALLBACK_ON_PICK},
	{"callback_on_drag", KnCALLBACK_ON_DRAG},
	{"callback_on_drop", KnCALLBACK_ON_DROP},
	{"callback_on_button", KnCALLBACK_ON_BUTTON},
	{"callback_always", KnCALLBACK_ALWAYS},

    };
    StringToLower((char *)from->addr, s);
    for(i = 0; i < XtNumber(names); i++) {
	if(!strcmp(s, names[i].name)) {
	    done(&(names[i].style), int);
	    return True;
	}
    }
    XtDisplayStringConversionWarning(dpy, (char *)from->addr,
				     XtRCallbackPolicy);
    return False;
}


void KnInteractorRealize(KnInteractor i, Widget w)
{
    XtTMRec tm;
    XtTranslations xlations;

    /* save current state and prevent for toolkit free */
    tm = w->core.tm;
    w->core.tm.translations = NULL;
    w->core.tm.proc_table = NULL;
    
    xlations = i->tm.translations;
    i->tm.translations = NULL;
    
    XtVaSetValues(w, XtNtranslations, xlations, 0);
    i->tm = w->core.tm;
    
    w->core.tm = tm;
}

/* init class fields and call super class Init proc */
void
KnInteractorInit(KnInteractor interactor, String translations, Widget w)
{
    memset(&(interactor->tm), 0, sizeof(XtTMRec));
    interactor->tm.translations = XtParseTranslationTable(translations);
    if(XtIsRealized(w)) {
	KnInteractorRealize(interactor, w);
    }
    else {
	/* delay to realize time */
	KnvasRegisterInteractor(w, interactor);
    }
}



/* allocates storage space for a new KnInteractor object */
KnInteractor
KnCreateInteractor(Widget w, String translations)
{  
    KnInteractor interactor;
    XtAppContext app_context;
    int actions_registered = False;

    app_context = XtWidgetToApplicationContext(w);
    if(!actions_registered) {
	XtAppAddActions(app_context, actions, XtNumber(actions));
	actions_registered = True;
    }
    interactor =  (KnInteractor)KlOMake(KnInteractorClass);
    KnInteractorInit(interactor, translations, w);
    return interactor;
}




KnInteractor
KnInteractorMerge(Widget w, KnInteractor i1, KnInteractor i2)
{
    XtTMRec tm;

    /* save current state and prevent from toolkit free */
    tm = w->core.tm;
    w->core.tm.translations = NULL;
    w->core.tm.proc_table = NULL;

    w->core.tm = i1->tm;
    XtOverrideTranslations(w, i2->tm.translations);
    i1->tm = w->core.tm;

    /* restore old state, preventing from free */
    w->core.tm = tm;
    return i1;
}




/* declare new   selectors
 */
void
KnInteractorImport()
{
	
}

void
KnInteractorRegisterConverters ()
{
    XtSetTypeConverter(XtRString, XtRCallbackPolicy, CvtStringToCallbackPolicy,
		       NULL, 0, XtCacheNone, NULL);
}


/* class initializations for KnInteractor objects
 * 
 * returns: 
 */
void
KnInteractorClassInitialize()
{
     /* declare  type */
    KlDeclareType(&KnInteractorClass, "KnInteractor", sizeof(struct KnInteractorStruct));

    /* declare methods */
    KlDeclareMethod(KnInteractorClass, KlSelMake, KnCreateInteractor);

    /* initializes standard interactors */
}







Hashtable interactorTable = NULL;

KnInteractor
KnCreateMoveInteractor(Widget knvas, char *name)
{
    static count = 0;
    char translations[1024];
    char ident[16];
    KnInteractor mi;
    struct KnInteractorStruct resources;

    /* this is a unique indentifier for the new interactor */
    sprintf(ident, "m%d", count++);
    
    XtGetSubresources(knvas, &resources, name, "KnMoveInteractor",
		      interactorResources, 
		      XtNumber(interactorResources), 0, 0);
    
    sprintf(translations, moveBehavior,
	    resources.button, ident,
	    resources.button, ident,
	    resources.button, ident);
    mi =  KnCreateInteractor(knvas, translations);
    mi->button = resources.button;
    mi->visibleSelection = resources.visibleSelection;
    mi->varyCursor = resources.varyCursor;
    mi->constraints = resources.constraints;
    mi->callbackPolicy = resources.callbackPolicy;

    /* register the new interactor in the sym table */
    if (NULL == interactorTable)
	interactorTable = HashtableMake(13);
    HashtableAdd(interactorTable, ident, mi);
    KlIncRef(mi);

    return mi;
}



KnInteractor
KnCreateResizeInteractor(Widget knvas, char *name)
{
    static count = 0;
    char translations[1024];
    char ident[16];
    KnInteractor mi;
    struct KnInteractorStruct resources;

    /* this is a unique indentifier for the new interactor */
    sprintf(ident, "r%d", count++);
    
    XtGetSubresources(knvas, &resources, name, "KnResizeInteractor",
		      interactorResources, 
		      XtNumber(interactorResources), 0, 0);
    
    sprintf(translations, resizeBehavior,
	    resources.button, ident,
	    resources.button, ident,
	    resources.button, ident);
    mi =  KnCreateInteractor(knvas, translations);
    mi->button = resources.button;
    mi->visibleSelection = resources.visibleSelection;
    mi->varyCursor = resources.varyCursor;
    mi->constraints = resources.constraints;
    mi->callbackPolicy = resources.callbackPolicy;

    /* register the new interactor in the sym table */
    if (NULL == interactorTable)
	interactorTable = HashtableMake(13);
    HashtableAdd(interactorTable, ident, mi);
    KlIncRef(mi);

    return mi;
}






KnInteractor
KnCreateSelectInteractor(Widget knvas, char *name)
{
    static count = 0;
    char translations[1024];
    char ident[16];
    KnInteractor mi;
    struct KnInteractorStruct resources;

    /* this is a unique indentifier for the new interactor */
    sprintf(ident, "s%d", count++);
    
    XtGetSubresources(knvas, &resources, name, "KnSelectInteractor",
		      interactorResources, 
		      XtNumber(interactorResources), 0, 0);
    
    sprintf(translations, selectBehavior,
	    resources.button, ident,
	    resources.button, ident,
	    resources.button, ident);
    mi =  KnCreateInteractor(knvas, translations);
    mi->button = resources.button;
    mi->visibleSelection = resources.visibleSelection;
    mi->varyCursor = resources.varyCursor;
    mi->constraints = resources.constraints;
    mi->callbackPolicy = resources.callbackPolicy;

    /* register the new interactor in the sym table */
    if (NULL == interactorTable)
	interactorTable = HashtableMake(13);
    HashtableAdd(interactorTable, ident, mi);
    KlIncRef(mi);

    return mi;
}

