/* Copyright 1994 GROUPE BULL -- See license conditions in file COPYRIGHT */
/* $Id: move.c,v 1.1.1.3 95/05/23 11:55:26 leon Exp $ */

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/cursorfont.h>
#include "misc.h"
#include "interactor.h"
#include "hash.h"



char 
moveBehavior[] = "\
 <Btn%dDown>: Select(%s)\n\
 <Btn%dMotion>: MoveDrag(%s)\n\
 <Btn%dUp>: MoveDrop(%s)\n\
";  



Transformer
GetGlobalTransformer(Widget w, KnO s)
{
    KnO p;
    KnvasWidget cw = (KnvasWidget)w;
    Transformer t;
    t = TransformerMake();
    p = s->parent;
    while(p) {
	if(p->t)
	    TransformerPremultiply(t, p->t);
	p = p->parent;
    }
    if(cw->knvas.t)
	TransformerPremultiply(t, cw->knvas.t);
    return t;
}


void
UnselectAll(Widget w)
{
    KnvasWidget cw = (KnvasWidget)w;
    Cardinal i;
    if(cw->knvas.sobjs->size > 0) {
	for(i = 0; i < cw->knvas.sobjs->size; i++){
	    KnUnselect(w, (KnO)cw->knvas.sobjs->list[i]);
	}
	/* destroys the old list and builds a new one */
	KlDecRef(cw->knvas.sobjs);
	cw->knvas.sobjs = KlListNMake(0);
	KlIncRef(cw->knvas.sobjs);
    }
}




 void
MoveDragProc(Widget w, XEvent *event, String *params, Cardinal *num)
{
    KnvasWidget cw = (KnvasWidget) w;
#define THIS cw->knvas
    InteractorCallbackStruct cs;
    KnInteractor mi;

    if(NULL == THIS.target) return;
    THIS.dispatched = True;
    
    if(1 == *num) {
	mi = HashtableFind(interactorTable, params[0]);
    }
    else {
	XtAppWarningMsg(XtWidgetToApplicationContext(w),
			invalidParameterCount,
			actionParameters,
			KnvasWarning,
			"Action 'MoveDragProc' requires 1 parameter",
			NULL, NULL);
	return;
    }
    
    cs.e = event;
    cs.interactor = mi;
    cs.target = THIS.target;
    cs.reason = KnCR_DRAG;
    if(0 == THIS.sobjs->size) 
	/* Move has no effect if there is no selected obj */
	return;
    if(0 == THIS.gobjs->size) {
	/* there is no ghost list, start move */
	Cardinal i;
	KnO kno, old;
	THIS.gt = (Transformer *)XtMalloc(sizeof(Transformer) * 
					  THIS.sobjs->size);
	/* create the ghost object list */
	for(i = 0; i < THIS.sobjs->size; i++) {	
	    old = (KnO)THIS.sobjs->list[i];
	    if(mi->callbackPolicy & KnCALLBACK_ON_DRAG) {
		KnCallCallbacks(w, old, XtNinteractorCallback, (XtPointer)&cs);
		XtCallCallbacks(w, XtNinteractorCallback, (XtPointer)&cs);
	    }
	    /* duplicates initial objects in the ghost list */
	    /* move the selected objects, sont that they may be drawn */
	    kno = (KnO)KlSend_copy(old);
	    KlListAppend(THIS.gobjs, kno);
	    THIS.gt[i] = GetGlobalTransformer(w, old);
	}
	THIS.firstx = THIS.lastx = event->xbutton.x;
	THIS.firsty = THIS.lasty = event->xbutton.y;
    }
    /*else*/{
	/* continue move */
	Cardinal i;
	KnO kno;
	Position bx, by, lx, ly;
	Position rbx, rby, rlx, rly;
	Position dx = 0, dy = 0;

	bx = event->xbutton.x;
	by = event->xbutton.y;
	lx = THIS.lastx;
	ly = THIS.lasty;
	for(i = 0; i < THIS.sobjs->size; i++) {
	    dx = dy = 0;
	    kno = (KnO)THIS.sobjs->list[i];
	    if(mi->callbackPolicy & KnCALLBACK_ON_DRAG) {
		KnCallCallbacks(w, kno, XtNinteractorCallback, (XtPointer)&cs);
		XtCallCallbacks(w, XtNinteractorCallback, (XtPointer)&cs);
	    }
	    TransformerInvTransform(THIS.gt[i], bx, by, &rbx, &rby);
	    TransformerInvTransform(THIS.gt[i], lx, ly, &rlx, &rly);
	    if(! mi->constraints || strchr(mi->constraints, 'h'))
		dx = rbx-rlx;
	    if(! mi->constraints || strchr(mi->constraints, 'v'))
		dy = rby-rly;
	    KnTranslate(w, kno, dx, dy);
	}
	THIS.lastx = event->xbutton.x;
	THIS.lasty = event->xbutton.y;
    }
    KnvasUpdateLayers(w);
#undef THIS
}




 void
MoveDropProc(Widget w, XEvent *event, String *params, Cardinal *num)
{
    KnvasWidget cw = (KnvasWidget) w;
#define THIS cw->knvas
    Cardinal i;    
    InteractorCallbackStruct cs;
    KnO initial, kno;
    Position bx, by, lx, ly, rbx, rby, rlx, rly;
	    Position x, y;
    KnInteractor mi;
    
    if(NULL == THIS.target) return;
    THIS.grab = NULL;		/* release grab */
    THIS.dispatched = True;
    
    if(1 == *num) {
	mi = HashtableFind(interactorTable, params[0]);
    }
    else {
	XtAppWarningMsg(XtWidgetToApplicationContext(w),
			invalidParameterCount,
			actionParameters,
			KnvasWarning,
			"Action 'MoveDropProc' requires 1 parameter",
			NULL, NULL);
	return;
    }
    bx = event->xbutton.x;
    by = event->xbutton.y;
    lx = THIS.lastx;
    ly = THIS.lasty;
    cs.e = event;
    cs.interactor = mi;
    cs.target = THIS.target;
    cs.reason = KnCR_DROP;
    cs.select = True;
    for(i = 0; i < THIS.gobjs->size; i++) {
	/* if gobjs is empty, an object may have been selected, but not moved,
           do not clear it: so we do the loop on gobjs */
	KlSend_clear((KnO)THIS.sobjs->list[i], w);
    }
    for(i = 0; i < THIS.gobjs->size; i++) {
	cs.select = True;
	kno = (KnO)THIS.sobjs->list[i];
	initial = (KnO)THIS.gobjs->list[i];
	cs.accept = True; 
	TransformerInvTransform(THIS.gt[i], bx, by, &rbx, &rby);
	TransformerInvTransform(THIS.gt[i], lx, ly, &rlx, &rly);
	/* the drop coordinates are the moved obj location */
	KlSend_position(kno, &x, &y);
	cs.tox = x;
	cs.toy = y;
	if(mi->callbackPolicy & KnCALLBACK_ON_DROP) {
	    KnCallCallbacks(w, kno, XtNinteractorCallback, (XtPointer)&cs);
	    XtCallCallbacks(w, XtNinteractorCallback, (XtPointer)&cs);
	}
	if(cs.accept) {
	    KnMove(w, kno, cs.tox, cs.toy);
	    KlSend_clear(kno,w );
	}
	else {
	    /* drop back to initial position */
	    KnTranslate(w, kno,
			    initial->x - kno->x,
			    initial->y - kno->y);
	    KlSend_clear(kno,w );
	}
	XtFree((void *)THIS.gt[i]);
    }
    if(!cs.select) {
	for(i = 0; i < THIS.sobjs->size; i++){
	    KnUnselect(w, (KnO)THIS.sobjs->list[i]);
	}
	KlDecRef(THIS.sobjs);
	THIS.sobjs = KlListNMake(0);
	KlIncRef(THIS.sobjs);
    }
    KnvasUpdateLayers(w);
    KlDecRef(THIS.gobjs);
    THIS.gobjs = KlListNMake(0);
    KlIncRef(THIS.gobjs);
    if(THIS.gt)
	XtFree((void *)THIS.gt);
    THIS.gt = NULL;
    if(mi->varyCursor) {
	XUndefineCursor(XtDisplay(w), XtWindow(w));
    }

#undef THIS
}



